/*{{{}}}*/
/*{{{  #includes*/
#define _POSIX_SOURCE   1
#define _POSIX_C_SOURCE 2

#include <assert.h>
#include <ctype.h>
#include <curses.h>
#include <limits.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cat.h"
#include "default.h"
#include "display.h"
#include "eval.h"
#include "misc.h"
#include "scanner.h"
#include "parser.h"
#include "sheet.h"
#include "wgetc.h"
/*}}}  */

/*{{{  variables*/
static WINDOW *edit;
/*}}}  */

/*{{{  do_edit*/
static void do_edit(Sheet *cursheet)
{
  /*{{{  variables*/
  char buf[1024];
  const char *s;
  int x,offx;
  Token **t;
  /*}}}  */
    
  if (locked(cursheet,cursheet->curx,cursheet->cury,cursheet->curz)) line_msg(edit,EDIT,ISLOCKED);
  else
  {
    offx=0;
    print(buf,sizeof(buf),0,1,getscientific(cursheet,cursheet->curx,cursheet->cury,cursheet->curz),DEF_REALPRECISION,getcont(cursheet,cursheet->curx,cursheet->cury,cursheet->curz));
    s=buf+strlen(buf);
    do
    {
      x=s-buf;
      if (line_edit(buf,sizeof(buf),edit,CONTENTS,&x,&offx)==-1) return;
      s=buf;
      t=scan(&s);
    } while (*s!='\0' && t==(Token**)0);
    if (t!=(Token**)0 && *t==(Token*)0) { free(t); t=(Token**)0; }
    putcont(cursheet,cursheet->curx,cursheet->cury,cursheet->curz,t);
  }
}
/*}}}  */
/*{{{  do_label*/
static void do_label(Sheet *sheet)
{
  /*{{{  variables*/
  char buf[1024];
  int x,offx,ok;
  Token t;
  /*}}}  */

  if (locked(sheet,sheet->curx,sheet->cury,sheet->curz)) line_msg(edit,DOLABEL,ISLOCKED);
  else
  {
    ok=x=offx=0;
    strcpy(buf,getlabel(sheet,sheet->curx,sheet->cury,sheet->curz));
    do
    {
      if (line_idedit(buf,sizeof(buf),edit,DOLABEL,&x,&offx)==-1) return;
      ok=((t=findlabel(sheet,buf)).type==EEK || (t.type==LOCATION && t.u.location[0]==sheet->curx && t.u.location[1]==sheet->cury && t.u.location[2]==sheet->curz));
      tfree(&t);
    } while (!ok);
    setlabel(sheet,sheet->curx,sheet->cury,sheet->curz,buf);
  }
}
/*}}}  */
/*{{{  do_columnwidth*/
static void do_columnwidth(Sheet *cursheet)
{
  /*{{{  variables*/
  int x,offx,n;
  /*}}}  */
            
  offx=0;
  x=0;
  n=columnwidth(cursheet,cursheet->curx,cursheet->curz);
  do if (line_numedit(&n,edit,COLWIDTH,&x,&offx)==-1) return; while (n<=0);
  setwidth(cursheet,cursheet->curx,cursheet->curz,n);
}
/*}}}  */
/*{{{  do_attribute*/
static void do_attribute(Sheet *cursheet)
{
  /*{{{  variables*/
  const char *menu[10];
  int c;
  int n;
  int offx;
  int x,y,z;
  int x1,y1,z1;
  int x2,y2,z2;
  /*}}}  */
      
  /*{{{  create menu*/
  menu[0]=strcpy(malloc(strlen(ADLEFT)+1),ADLEFT);
  menu[1]=strcpy(malloc(strlen(ADRIGHT)+1),ADRIGHT);
  menu[2]=strcpy(malloc(strlen(ADCENTERED)+1),ADCENTERED);
  menu[3]=strcpy(malloc(strlen(DOSCI)+1),DOSCI);
  menu[4]=strcpy(malloc(strlen(NOSCI)+1),NOSCI);
  menu[5]=strcpy(malloc(strlen(PRECISION)+1),PRECISION);
  menu[6]=strcpy(malloc(strlen(SHADOW)+1),SHADOW);
  menu[7]=strcpy(malloc(strlen(LABEL)+1),LABEL);
  menu[8]=strcpy(malloc(strlen(DOLOCK)+1),DOLOCK);
  menu[9]=(const char*)0;
  /*}}}  */
  if (cursheet->mark1x==-1)
  /*{{{  range is the current cell*/
  {
    x1=x2=cursheet->curx;
    y1=y2=cursheet->cury;
    z1=z2=cursheet->curz;
  }
  /*}}}  */
  else
  /*{{{  range is the marked cube*/
  {
    x1=cursheet->mark1x; x2=cursheet->mark2x; order(&x1,&x2);
    y1=cursheet->mark1y; y2=cursheet->mark2y; order(&y1,&y2);
    z1=cursheet->mark1z; z2=cursheet->mark2z; order(&z1,&z2);
  }
  /*}}}  */
  if ((c=line_menu(edit,cursheet->mark1x==-1 ? CELLATTR : BLOCKATTR,menu,getadjust(cursheet,x1,y1,z1)))!=-1)
  {
    if (cursheet->mark1x==-1 && c<=7 && locked(cursheet,cursheet->curx,cursheet->cury,cursheet->curz)) line_msg(edit,CELLATTR,ISLOCKED);
    else
    {
      if (c==5)
      {
        offx=0;
        x=0;
        n=getprecision(cursheet,x1,y1,z1);
        do if (line_numedit(&n,edit,cursheet->mark1x==-1 ? PRECCELL : PRECBLOCK,&x,&offx)==-1) return; while (n<=0 || n>20);
      }
      else if (c==6)
      {
        if ((n=line_ok(edit,cursheet->mark1x==-1 ? SHADOWCELL : SHADOWCUBE,shadowed(cursheet,x1,y1,z1)))==-1) return;
      }
      else if (c==7)
      {
        do_label(cursheet);
      }
      else if (c==8)
      {
        if ((n=line_ok(edit,cursheet->mark1x==-1 ? LOCKCELL : LOCKBLOCK,locked(cursheet,x1,y1,z1)))==-1) return;
      }    
      if (c!=7) for (x=x1; x<=x2; ++x) for (y=y1; y<=y2; ++y) for (z=z1; z<=z2; ++z) switch (c)
      {
        /*{{{  0       -- adjust left*/
        case 0: setadjust(cursheet,x,y,z,LEFT); break;
        /*}}}  */
        /*{{{  1       -- adjust right*/
        case 1: setadjust(cursheet,x,y,z,RIGHT); break;
        /*}}}  */
        /*{{{  2       -- adjust centered*/
        case 2: setadjust(cursheet,x,y,z,CENTER); break;
        /*}}}  */
        /*{{{  3       -- set scientific notation flag*/
        case 3: setscientific(cursheet,x,y,z,1); break;
        /*}}}  */
        /*{{{  4       -- clear scientific notation flag*/
        case 4: setscientific(cursheet,x,y,z,0); break;
        /*}}}  */
        /*{{{  5       -- set precision*/
        case 5: setprecision(cursheet,x,y,z,n); break;
        /*}}}  */
        /*{{{  6       -- shadow*/
        case 6:
        {
          int rx;

          if (n==0) for (rx=x+1; shadowed(cursheet,rx,y,z); ++rx) shadow(cursheet,rx,y,z,0);
          else if (x>0) shadow(cursheet,x,y,z,1);
          break;
        }
        /*}}}  */
        /*{{{  8       -- lock*/
        case 8: lockcell(cursheet,x,y,z,n); break;
        /*}}}  */
        /*{{{  default -- should not happen*/
        default: assert(0);
        /*}}}  */
      }
      if (c==6 && x1==0 && n==1) line_msg(edit,SHADOWCELL,SHADOWBOO);
    }
  }
  /*{{{  free menu*/
  free((char*)menu[0]);
  free((char*)menu[1]);
  free((char*)menu[2]);
  free((char*)menu[3]);
  free((char*)menu[4]);
  free((char*)menu[5]);
  free((char*)menu[6]);
  free((char*)menu[7]);
  free((char*)menu[8]);
  /*}}}  */
  cursheet->marking=0;
  cursheet->mark1x=-1;
  forceupdate(cursheet);
}
/*}}}  */
/*{{{  do_savexdr*/
static void do_savexdr(Sheet *cursheet)
{
  char buf[_POSIX_PATH_MAX];
  const char *msg;
      
  if (cursheet->name==(char*)0)
  {
    /*{{{  variables*/
    int x,offx;
    /*}}}  */
        
    x=0;
    offx=0;
    buf[0]='\0';
    if (line_edit(buf,sizeof(buf),edit,SAVESHEET,&x,&offx)==-1) return;
    cursheet->name=strcpy(malloc(strlen(buf)+1),buf);
  }
  if ((msg=savexdr(cursheet,cursheet->name))!=(char*)0) line_msg(edit,SAVESHEET,msg);
}
/*}}}  */
/*{{{  do_savetbl*/
static void do_savetbl(Sheet *cursheet)
{
  /*{{{  variables*/
  char buf[_POSIX_PATH_MAX+4];
  char pages[80];
  const char *msg;
  int x,offx;
  /*}}}  */
      
  if (cursheet->name==(char*)0) buf[0]='\0';
  else
  {
    strcpy(buf,cursheet->name);
    strcat(buf,".tbl");
  }
  x=0;
  offx=0;
  if (line_edit(buf,sizeof(buf),edit,SAVETBLF,&x,&offx)==-1) return;
  if (buf[0]=='\0') return;
  pages[0]='\0';
  x=0;
  offx=0;
  if (line_edit(pages,sizeof(pages),edit,PAGES,&x,&offx)==-1) return;  
  if ((msg=savetbl(cursheet,buf,pages))!=(char*)0) line_msg(edit,SAVETBLF,msg);
}
/*}}}  */
/*{{{  do_savetext*/
static void do_savetext(Sheet *cursheet)
{
  /*{{{  variables*/
  char buf[_POSIX_PATH_MAX+4];
  char pages[80];
  const char *msg;
  int x,offx;
  /*}}}  */
      
  if (cursheet->name==(char*)0) buf[0]='\0';
  else
  {
    strcpy(buf,cursheet->name);
    strcat(buf,".doc");
  }
  x=0;
  offx=0;
  if (line_edit(buf,sizeof(buf),edit,SAVETEXTF,&x,&offx)==-1) return;
  if (buf[0]=='\0') return;
  pages[0]='\0';
  x=0;
  offx=0;
  if (line_edit(pages,sizeof(pages),edit,PAGES,&x,&offx)==-1) return;  
  if ((msg=savetext(cursheet,buf,pages))!=(char*)0) line_msg(edit,SAVETEXTF,msg);
}
/*}}}  */
/*{{{  do_loadxdr*/
static void do_loadxdr(Sheet *cursheet)
{
  /*{{{  variables*/
  int x,offx;
  char buf[_POSIX_PATH_MAX];
  const char *msg;
  /*}}}  */
      
  x=0;
  offx=0;
  buf[0]='\0';
  while (buf[0]=='\0') if (line_edit(buf,sizeof(buf),edit,LOADSHEET,&x,&offx)==-1) return;
  if (cursheet->name!=(char*)0) free(cursheet->name);
  cursheet->name=strcpy(malloc(strlen(buf)+1),buf);
  if ((msg=loadxdr(cursheet,cursheet->name))!=(const char*)0) line_msg(edit,LOADSHEET,msg);
}
/*}}}  */
/*{{{  do_mark*/
static void do_mark(Sheet *cursheet)
{
  if (!cursheet->marking && cursheet->mark1x==-1)
  {
    cursheet->mark1x=cursheet->mark2x=cursheet->curx;
    cursheet->mark1y=cursheet->mark2y=cursheet->cury;
    cursheet->mark1z=cursheet->mark2z=cursheet->curz;
    cursheet->marking=1;
  }
  else if (cursheet->marking) cursheet->marking=0;
  else cursheet->mark1x=-1;
}
/*}}}  */
/*{{{  do_save*/
static void do_save(Sheet *cursheet)
{
  /*{{{  variables*/
  const char *menu[]={ (const char*)0, (const char*)0, (const char*)0, (const char*)0 };
  /*}}}  */
  
  menu[0]=strcpy(malloc(strlen(ASXDR)+1),ASXDR);
  menu[1]=strcpy(malloc(strlen(TBL)+1),TBL);
  menu[2]=strcpy(malloc(strlen(SAVETEXT)+1),SAVETEXT);  
  switch (line_menu(edit,SAVEMENU,menu,0))
  {
    /*{{{  -1      -- abort*/
    case -1: break;
    /*}}}  */
    /*{{{   0      -- save in XDR*/
    case 0: do_savexdr(cursheet); break;
    /*}}}  */
    /*{{{   1      -- save in tbl format*/
    case 1: do_savetbl(cursheet); break;
    /*}}}  */
    /*{{{   2      -- save in text format*/
    case 2: do_savetext(cursheet); break;
    /*}}}  */
    /*{{{  default -- should not happen*/
    default: assert(0);
    /*}}}  */
  }
  free((char*)menu[0]);
  free((char*)menu[1]);
  free((char*)menu[2]);
}
/*}}}  */
/*{{{  do_clear*/
static void do_clear(Sheet *sheet)
{
  /*{{{  variables*/
  int x,y,z;
  int x1,y1,z1;
  int x2,y2,z2;
  /*}}}  */

  if (sheet->mark1x==-1 && locked(sheet,sheet->curx,sheet->cury,sheet->curz)) line_msg(edit,CLEARCELL,ISLOCKED);
  else
  {
    if (sheet->mark1x!=-1) if (line_ok(edit,CLEARBLOCK,0)!=1) return;
    if (sheet->mark1x==-1)
    /*{{{  range is the current cell*/
    {
      x1=x2=sheet->curx;
      y1=y2=sheet->cury;
      z1=z2=sheet->curz;
    }
    /*}}}  */
    else
    /*{{{  range is the marked cube*/
    {
      x1=sheet->mark1x; x2=sheet->mark2x; order(&x1,&x2);
      y1=sheet->mark1y; y2=sheet->mark2y; order(&y1,&y2);
      z1=sheet->mark1z; z2=sheet->mark2z; order(&z1,&z2);
    }
    /*}}}  */
    for (x=x1; x<=x2; ++x) for (y=y1; y<=y2; ++y) for (z=z1; z<=z2; ++z) freecell(sheet,x,y,z);
    forceupdate(sheet);
    sheet->marking=0;
    sheet->mark1x=-1;
  }
}
/*}}}  */
/*{{{  do_insert*/
static void do_insert(Sheet *sheet)
{
  /*{{{  variables*/
  int x1,y1,z1,x2,y2,z2;
  const char *menu[]={ (const char*)0, (const char*)0, (const char*)0, (const char*)0 };
  /*}}}  */
  
  menu[0]=strcpy(malloc(strlen(INX)+1),INX);
  menu[1]=strcpy(malloc(strlen(INY)+1),INY);
  menu[2]=strcpy(malloc(strlen(INZ)+1),INZ);  
  if (sheet->mark1x==-1)
  /*{{{  range is the current cell*/
  {
    x1=x2=sheet->curx;
    y1=y2=sheet->cury;
    z1=z2=sheet->curz;
  }
  /*}}}  */
  else
  /*{{{  range is the marked cube*/
  {
    x1=sheet->mark1x; x2=sheet->mark2x; order(&x1,&x2);
    y1=sheet->mark1y; y2=sheet->mark2y; order(&y1,&y2);
    z1=sheet->mark1z; z2=sheet->mark2z; order(&z1,&z2);
  }
  /*}}}  */
  switch(line_menu(edit,sheet->mark1x==-1 ? INSCELL : INSCUBE,menu,0))
  {
    /*{{{   0      -- columns*/
    case 0: insertcube(sheet,x1,y1,z1,x2,y2,z2,IN_X); break;
    /*}}}  */
    /*{{{   1      -- rows*/
    case 1: insertcube(sheet,x1,y1,z1,x2,y2,z2,IN_Y); break;
    /*}}}  */
    /*{{{   2      -- depth*/
    case 2: insertcube(sheet,x1,y1,z1,x2,y2,z2,IN_Z); break;
    /*}}}  */
    /*{{{  -1      -- abort*/
    case -1: goto greak;
    /*}}}  */
    /*{{{  default -- should not happen*/
    default: assert(0);
    /*}}}  */
  }
  sheet->marking=0;
  sheet->mark1x=-1;
  greak:
  free((char*)menu[0]);
  free((char*)menu[1]);
  free((char*)menu[2]);
}
/*}}}  */
/*{{{  do_delete*/
static void do_delete(Sheet *sheet)
{
  /*{{{  variables*/
  int x1,y1,z1,x2,y2,z2;
  const char *menu[]={ (const char*)0, (const char*)0, (const char*)0, (const char*)0 };
  /*}}}  */
  
  menu[0]=strcpy(malloc(strlen(INX)+1),INX);
  menu[1]=strcpy(malloc(strlen(INY)+1),INY);
  menu[2]=strcpy(malloc(strlen(INZ)+1),INZ);  
  if (sheet->mark1x==-1)
  /*{{{  range is the current cell*/
  {
    x1=x2=sheet->curx;
    y1=y2=sheet->cury;
    z1=z2=sheet->curz;
  }
  /*}}}  */
  else
  /*{{{  range is the marked cube*/
  {
    x1=sheet->mark1x; x2=sheet->mark2x; order(&x1,&x2);
    y1=sheet->mark1y; y2=sheet->mark2y; order(&y1,&y2);
    z1=sheet->mark1z; z2=sheet->mark2z; order(&z1,&z2);
  }
  /*}}}  */
  switch(line_menu(edit,sheet->mark1x==-1 ? DELCELL : DELBLOCK,menu,0))
  {
    /*{{{   0      -- columns*/
    case 0: deletecube(sheet,x1,y1,z1,x2,y2,z2,IN_X); break;
    /*}}}  */
    /*{{{   1      -- rows*/
    case 1: deletecube(sheet,x1,y1,z1,x2,y2,z2,IN_Y); break;
    /*}}}  */
    /*{{{   2      -- depth*/
    case 2: deletecube(sheet,x1,y1,z1,x2,y2,z2,IN_Z); break;
    /*}}}  */
    /*{{{  -1      -- abort*/
    case -1: goto greak;
    /*}}}  */
    /*{{{  default -- should not happen*/
    default: assert(0);
    /*}}}  */
  }
  forceupdate(sheet);
  sheet->marking=0;
  sheet->mark1x=-1;
  greak:
  free((char*)menu[0]);
  free((char*)menu[1]);
  free((char*)menu[2]);
}
/*}}}  */
/*{{{  do_about*/
static void do_about(Sheet *sheet)
{
  line_msg(edit,(const char*)0,"Version 0.1, Copyright (c) 1995 by Michael Haardt");
}
/*}}}  */
/*{{{  do_move*/
static void do_move(Sheet *sheet, int copy)
{
  if (sheet->mark1x==-1) line_msg(edit,copy ? COPYBLOCK : MOVEBLOCK,NOMARK);
  else if (line_ok(edit,copy ? COPYBLOCK : MOVEBLOCK,0)==1)
  {
    int x1,y1,z1;
    int x2,y2,z2;

    x1=sheet->mark1x; x2=sheet->mark2x; order(&x1,&x2);
    y1=sheet->mark1y; y2=sheet->mark2y; order(&y1,&y2);
    z1=sheet->mark1z; z2=sheet->mark2z; order(&z1,&z2);
    moveblock(sheet,x1,y1,z1,x2,y2,z2,sheet->curx,sheet->cury,sheet->curz,copy);
    sheet->marking=0;
    sheet->mark1x=-1;    
  }
}
/*}}}  */
/*{{{  leaveanyway*/
static int leaveanyway(Sheet *sheet)
{
  int result;
  
  if (sheet->changed)
  {
    result=line_ok(edit,LEAVE,0);
    if (result==-1) return 0; else return result;
  }
  return 1;
}
/*}}}  */

/*{{{  work*/
static int work(Sheet *cursheet, chtype c)
{
  switch (c)
  {
    /*{{{  UP*/
    case KEY_UP:
    {
      if (cursheet->cury>0) --cursheet->cury;
      if (cursheet->marking) cursheet->mark2y=cursheet->cury;
      break;
    }
    /*}}}  */
    /*{{{  DOWN*/
    case KEY_DOWN:
    {
      ++cursheet->cury;
      if (cursheet->marking) cursheet->mark2y=cursheet->cury;
      break;
    }
    /*}}}  */
    /*{{{  LEFT*/
    case KEY_LEFT:
    {
      if (cursheet->curx>0) --cursheet->curx;
      if (cursheet->marking) cursheet->mark2x=cursheet->curx;
      break;
    }
    /*}}}  */
    /*{{{  RIGHT*/
    case KEY_RIGHT:
    {
      ++cursheet->curx; while (shadowed(cursheet,cursheet->curx,cursheet->cury,cursheet->curz)) ++cursheet->curx;
      if (cursheet->marking) cursheet->mark2x=cursheet->curx;
      break;
    }
    /*}}}  */
    /*{{{  <*/
    case '<':
    {
      cursheet->cury=0;
      if (cursheet->marking) cursheet->mark2y=cursheet->cury;
      break;
    }
    /*}}}  */
    /*{{{  >*/
    case '>':
    {
      cursheet->cury=(cursheet->dimy ? cursheet->dimy-1 : 0);
      if (cursheet->marking) cursheet->mark2y=cursheet->cury;
      break;
    }
    /*}}}  */
    /*{{{  BEG*/
    case KEY_BEG:
    {
      cursheet->curx=0;
      if (cursheet->marking) cursheet->mark2x=cursheet->curx;
      break;
    }
    /*}}}  */
    /*{{{  END*/
    case KEY_END:
    {
      cursheet->curx=(cursheet->dimx ? cursheet->dimx-1 : 0);
      while (shadowed(cursheet,cursheet->curx,cursheet->cury,cursheet->curz)) ++cursheet->curx;
      if (cursheet->marking) cursheet->mark2x=cursheet->curx;
      break;
    }
    /*}}}  */
    /*{{{  +      */
    case '+':
    {
      ++cursheet->curz; 
      if (cursheet->marking) cursheet->mark2z=cursheet->curz;
      break;
    }
    /*}}}  */
    /*{{{  -    */
    case '-':
    {
      if (cursheet->curz>0) --cursheet->curz; 
      if (cursheet->marking) cursheet->mark2z=cursheet->curz;
      break;
    }
    /*}}}  */
    /*{{{  **/
    case '*':
    {
      cursheet->curz=(cursheet->dimz ? cursheet->dimz-1 : 0);
      if (cursheet->marking) cursheet->mark2z=cursheet->curz;
      break;
    }
    /*}}}  */
    /*{{{  _    */
    case '_':
    {
      cursheet->curz=0;
      if (cursheet->marking) cursheet->mark2z=cursheet->curz;
      break;
    }
    /*}}}  */
    /*{{{  e*/
    case 'e': do_edit(cursheet); break;
    /*}}}  */
    /*{{{  s*/
    case 's': saveport(cursheet,"foo.port"); break;
    /*}}}  */
    /*{{{  l*/
    case 'l':
    {
    const char *err;

    if ((err=loadport(cursheet,"foo.port"))!=(const char*)0) line_msg(edit,"load port",err);
    break;
    }
    /*}}}  */
    /*{{{  :*/
    case ':':
    {
      WINDOW *foo;
      chtype *background;

      foo=newpopupwin(10,10,10,10,&background);
      wclear(foo);
      wrefresh(foo);
      wgetc(foo);
      delpopupwin(foo,&background);
      break;
    }
    /*}}}  */
    /*{{{  .*/
    case KEY_MARK:
    case '.': do_mark(cursheet); break;
    /*}}}  */
    /*{{{  ?*/
    case '?':
    {
      /*{{{  variables*/
      const char *menu[12];
      /*}}}  */

      menu[0]=strcpy(malloc(strlen(ATTR)+1),ATTR);
      menu[1]=strcpy(malloc(strlen(CWIDTH)+1),CWIDTH);
      menu[2]=strcpy(malloc(strlen(CLEAR)+1),CLEAR);
      menu[3]=strcpy(malloc(strlen(INSERT)+1),INSERT);
      menu[4]=strcpy(malloc(strlen(DELETE)+1),DELETE);
      menu[5]=strcpy(malloc(strlen(MOVE)+1),MOVE);
      menu[6]=strcpy(malloc(strlen(COPY)+1),COPY);  
      menu[7]=strcpy(malloc(strlen(LOAD)+1),LOAD);
      menu[8]=strcpy(malloc(strlen(SAVE)+1),SAVE);
      menu[9]=strcpy(malloc(strlen(ABOUT)+1),ABOUT);
      menu[10]=strcpy(malloc(strlen(QUIT)+1),QUIT);
      menu[11]=(const char*)0;
      switch (line_menu(edit,MAINMENU,menu,0))
      {
        case -1: break;
        case 0: do_attribute(cursheet); break;
        case 1: do_columnwidth(cursheet); break;
        case 2: do_clear(cursheet); break;
        case 3: do_insert(cursheet); break;
        case 4: do_delete(cursheet); break;
        case 5: do_move(cursheet,0); break;
        case 6: do_move(cursheet,1); break;
        case 7: do_loadxdr(cursheet); break;
        case 8: do_save(cursheet); break;
        case 9: do_about(cursheet); break;
        case 10: return 1; break;
        default: assert(0);
      }
      free((char*)menu[0]);
      free((char*)menu[1]);
      free((char*)menu[2]);
      free((char*)menu[3]);
      free((char*)menu[4]);
      free((char*)menu[5]);
      free((char*)menu[6]);
      free((char*)menu[7]);
      free((char*)menu[8]);
      free((char*)menu[9]);
      free((char*)menu[10]);
      break;
    }
    /*}}}  */
    /*{{{  q*/
    case 'q': return 1;
    /*}}}  */
  }
  return 0;
}
/*}}}  */
                  
/*{{{  main*/
int main(int argc, char *argv[], char *env[])
{
  chtype c;
  Sheet sheet,*cursheet;

  setlocale(LC_ALL,"");
  initscr();
  curs_set(0);
  noecho();
  raw();
  keypad(stdscr,1);
  cursheet=&sheet;
  cursheet->curx=cursheet->cury=cursheet->curz=0;
  cursheet->offx=cursheet->offy=0;
  cursheet->dimx=cursheet->dimy=cursheet->dimz=0;
  cursheet->sheet=(Cell**)0;
  cursheet->column=(int*)0;
  cursheet->window=newwin(LINES-1,COLS,0,0);
  cursheet->name=(char*)0;
  cursheet->mark1x=-1;
  cursheet->marking=0;
  cursheet->changed=0;
  if (argc>=2)
  {
    const char *msg;

    cursheet->name=strcpy(malloc(strlen(argv[1])+1),argv[1]);
    if ((msg=loadxdr(cursheet,cursheet->name))!=(const char*)0) line_msg(edit,LOADSHEET,msg);
  }
  edit=newwin(1,COLS,LINES-1,0);
  keypad(edit,1);
  wmove(edit,0,0); wrefresh(edit); /* does not work else?! */
  line_msg(edit,(const char*)0,TOMENU);
  do
  {
    redraw_sheet(cursheet);
    c=wgetc(stdscr);
    wmove(edit,0,0); wclrtoeol(edit); wrefresh(edit);
  } while (work(cursheet,c)==0 || leaveanyway(cursheet)==0);    
  curs_set(1);
  echo();
  noraw();
  endwin();
  return 0;
}
/*}}}  */
